"""Set of tests for feature_selector module."""

from AutoCarver.feature_selector import *
from pytest import fixture


def test_feature_selector_quanti():
    assert False, "TODO"

def test_feature_selector_quali():
    assert False, "TODO"
